﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace CarBuilder
{
    public partial class MainForm : Form
    {
        public MainForm()
        {
            InitializeComponent();
        }

        private void availablePartsView_ItemDrag(object sender, ItemDragEventArgs e) //Inicializacia drag&drop.
        {
            availablePartsView.DoDragDrop(availablePartsView.SelectedItems, DragDropEffects.All); //Drag SelectedItems.
        }

        private void usedPartsView_DragOver(object sender, DragEventArgs e) //Toto je len eyecandy.
        {
            if (e.Data.GetDataPresent(typeof(ListView.SelectedListViewItemCollection))) //Zistujeme ci tahame spravny objekt.
            {
                e.Effect = DragDropEffects.Copy; //Zmeni kurzor na drag&drop copy kurzor ked hoverujeme s tahanym objektom kompatibilneho typu.
            }
        }

        private void usedPartsView_DragDrop(object sender, DragEventArgs e) //Drop
        {
            ListView.SelectedListViewItemCollection items = (ListView.SelectedListViewItemCollection)e.Data.GetData(typeof(ListView.SelectedListViewItemCollection)); //Ziskanie tahaneho objektu.

            foreach (ListViewItem item in items)
            {
                usedPartsView.Items.Add((ListViewItem)item.Clone()); //Kazdy prvok ktory sme tahali (mozeme oznacit viac naraz) skopirujeme.
            }
        }

        private void usedPartsView_KeyDown(object sender, KeyEventArgs e) //Mazanie vybranych suciastok deletom.
        {
            if (e.KeyCode == Keys.Delete)
            {
                foreach (ListViewItem item in usedPartsView.SelectedItems)
                {
                    usedPartsView.Items.Remove(item);
                }
            }
        }

        private void assambleButton_Click(object sender, EventArgs e) //Zakladne pouzitie vlakien.
        {
            Task.Factory.StartNew(Assemble); //Tot vse. Proste spustime novu task, ktora vykona metodu Assemble(). Spustam cez Task.Factory aby som vyuzil tzv. "Thread Pooling" (neznalci at googli).
        }

        private void Assemble()
        {
            ListViewItem[] parts = InitAssembly(); //Inicializacia

            int totalCost = 0;

            foreach (ListViewItem part in parts)
            {
                totalCost += Assemble(part); //Postupne namontujeme kazdu suciastku
            }

            EndAssembly(totalCost); //Upratanie
        }

        private ListViewItem[] InitAssembly()
        {
            ListViewItem[] parts = null;

            Invoke(new Action(() => //Inicializacia GUI. Pristupovat ku GUI z ineho ako GUI vlakna je zakazane, takze je potrebne pouzit specialnu metodu Invoke, ktora bere obecneho delegata.
                                    //Kedze sa jedna o obecneho delegata (je jedno ake a kolko parametrov ma) je potrebne pouzit konstruktor new Action(...).
                                    //V tomto pripade som ... nahradil bezparametrickou lambda abstrakciou
                {
                    parts = new ListViewItem[usedPartsView.Items.Count]; //montovane suciastky su natiahneme do tohto vlakna.
                    usedPartsView.Items.CopyTo(parts, 0);

                    statusLabel.Visible = true;

                    progressBar.Maximum = usedPartsView.Items.Count;
                    progressBar.Value = 0;
                    progressBar.Visible = true;

                    assambleButton.Enabled = false; //Nechce sa mi prerusovat paralelne ulohy a nema zmysel, aby sa suciastky montovali viackrat naraz... tak proste zablokujem tlacitka -.-'
                    expressButton.Enabled = false;
                }));

            return parts;
        }

        private int Assemble(ListViewItem part)
        {
            int time = (Int32.Parse(part.SubItems[1].Text) * 100);
            Thread.Sleep(time); //Simulacia doby montaze. Vlakno proste zaspi na dany cas. GUI vlakno vsak bude nadalej reagovat.
                                //(Obycajne nebudete pouzivat Thread.Sleep() ale budete vo vlaknach pocitat zlozite vypocty.)

            Invoke(new Action(() => //Ked skonci montaz jednej zo suciastok ukazeme to na progress bare.
            {
                progressBar.Value++;
            }));

            return Int32.Parse(part.SubItems[2].Text); //Vrat cenu suciastky a montaze.
        }

        private void EndAssembly(int totalCost)
        {
            Invoke(new Action<int>(cost => //Nakoniec po sebe upraceme GUI a vypiseme spravu o uspechu.
            {
                statusLabel.Visible = false;
                progressBar.Visible = false;

                assambleButton.Enabled = true;
                expressButton.Enabled = true;

                usedPartsView.Items.Clear();
                MessageBox.Show(String.Format(@"The modifications to the car have been successfully completed!
The total cost amounts to: {0}", cost), "Success!");
            }), new object[] { totalCost }); //Predavanie parametrov delegatu, ktory sa ma vykonat na GUI vlakne. Narozdiel od predchodzich volani Invoke() tentoraz uz moj delegat berie int.
        }

        private void expressButton_Click(object sender, EventArgs e)
        {
            Task.Factory.StartNew(ExpressAssemble); //Opat volanie v novom vlakne.
        }

        private void ExpressAssemble()
        {
            ListViewItem[] parts = InitAssembly();

            int totalCost = 100000;

            Task[] tasks = parts //Narozdiel ok klasickej montaze, expresna montaz montuje vsetky suciasky naraz (paralelne), tudiz kazdu v inom vlakne.
                .Select(part => Task.Factory.StartNew(() => //Na zadanie ulohy vlaknu tentoraz pouzivam lambda abstrakciu.
                {
                    int cost = Assemble(part);

                    lock (parts) //Synchronizacia pristupu k premennej totalCost. Zamykam cez parts, kedze totalCost je int - hodnotovy typ.
                    {
                        totalCost += cost;
                    }
                }))
                .ToArray();

            Task.WaitAll(tasks); //Pockame kym skonci montaz vsetkych suciastok.

            EndAssembly(totalCost);
        }
    }
}
